#ifndef ___LCD_C
#define ___LCD_C

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <vdr/config.h>
#include <vdr/tools.h>
#include <vdr/remote.h>
#include "setup.h"
#include "lcd.h"
#include "sockets.h"
#include "i18n.h"
#include <vdr/plugin.h>
#include <vdr/timers.h>
#include <vdr/channels.h>

#ifdef LCD_EXT_KEY_CONF 
#include LCD_EXT_KEY_CONF 
#else 
#include "lcdkeyconf.h"
#endif

// character mapping for output, see cLcd::Write
#include "lcdtranstbl.h"  

#define LCDMENU   0
#define LCDTITLE  1
#define LCDREPLAY 2
#define LCDMISC   3
#define LCDVOL    4

/**
 * Sets the "output state" for the device. We use this to control the icons around the outside the
 * display. The bits in \c state correspond to the icons as follows:
 *
 * bit 0       : disc icon (0=off, 1='spin') , if Toprow==4, use CD-animation, else use "HDD-recording-animation"
 * bit 1,2,3   : top row (0=none, 1=music, 2=movie, 3=photo, 4=CD/DVD, 5=TV, 6=Web, 7=News/Weather)
 * bit 4,5     : 'speaker' icons (0=off, 1=L+R, 2=5.1ch, 3=7.1ch)
 * bit 6       : S/PDIF icon
 * bit 7       : 'SRC'
 * bit 8       : 'FIT'
 * bit 9       : 'TV'
 * bit 10      : 'HDTV'
 * bit 11      : 'SRC1'
 * bit 12      : 'SRC2'
 * bit 13,14,15: bottom-right icons (0=off, 1=MP3, 2=OGG, 3=WMA, 4=WAV)
 * bit 16,17,18: bottom-middle icons (0=off, 1=MPG, 2=AC3, 3=DTS, 4=WMA)
 * bit 19,20,21: bottom-left icons (0=off, 1=MPG, 2=DIVX, 3=XVID, 4=WMV)
 * bit 22      : 'VOL' (volume)
 * bit 23      : 'TIME'
 * bit 24      : 'ALARM'
 * bit 25      : 'REC' (recording)
 * bit 26      : 'REP' (repeat)
 * bit 27      : 'SFL' (shuffle)
 * bit 28      : Abuse this for progress bars (if set to 1), lower bits represent
 *               the length (6 bits each: P|6xTP|6xTL|6xBL|6xBP with P = bit 28,
 *               TP=Top Progress, TL = Top Line, BL = Bottom Line, BP = Bottom Progress).
 *               If bit 28 is set to 1, lower bits are interpreted as
 *               lengths; otherwise setting the symbols as usual.
 *               0 <= length <= 32, bars extend from left to right.
 *               length > 32, bars extend from right to left, length is counted
 *               from 32 up (i.e. 35 means a length of 3).
 *                       
 *     Remember: There are two kinds of calls!
 *               With bit 28 set to 1: Set all bars (leaving the symbols as is),
 *               with bit 28 set to 0: Set the symbols (leaving the bars as is).
 *     Beware:   TODO: May become a race condition, if both calls are executed
 *                     before the display gets updated. Keep this in mind in your
 *                     client-code.
 * bit 29      : 'disc-in icon' - half ellipsoid under the disc symbols (0=off, 1=on)
 */
#define IMON_OUTPUT_DISC        0
#define IMON_OUTPUT_TOP_ROW     1
#define IMON_OUTPUT_SPKR        4
#define IMON_OUTPUT_SPDIF       6
#define IMON_OUTPUT_SRC         7
#define IMON_OUTPUT_FIT         8
#define IMON_OUTPUT_TV          9
#define IMON_OUTPUT_HDTV        10
#define IMON_OUTPUT_SCR1        11
#define IMON_OUTPUT_SCR2        12
#define IMON_OUTPUT_BR_ICONS    13
#define IMON_OUTPUT_BM_ICONS    16
#define IMON_OUTPUT_BL_ICONS    19
#define IMON_OUTPUT_VOL         22
#define IMON_OUTPUT_TIME        23
#define IMON_OUTPUT_ALARM       24
#define IMON_OUTPUT_REC         25
#define IMON_OUTPUT_REP         26
#define IMON_OUTPUT_SFL         27
#define IMON_OUTPUT_PBARS       28
#define IMON_OUTPUT_DISK_IN     29

// unfortunalety, I can put any number (1-4) here and the bars always begin at the
// bottom of the display (on my 4 lines display - a Matrix Orbital 20x4 ).
// So there's no harm setting this to 2 for 2 lines displays.
//#define SA_Y 2

// public:

cLcd::cLcd() {
    int i,j;
    channelSwitched = false;
//    dimmed = false;
    connected=false;
    ThreadStateData.showvolume=false;
    ThreadStateData.newscroll=false;
    sock=wid=hgt=cellwid=cellhgt=0;
    closing = false;
    ThreadStateData.outputValue = 0;
    ThreadStateData.progressBarsValue = 0;
    femonReceiver = NULL;
    attachFemonReceiver = false;
    SummaryText = NULL;
    
    replayDvbApi=NULL;
    primaryDvbApi=NULL;

    for (i=0;i<LCDMAXSTATES;i++) for (j=0;j<4;j++) {
        ThreadStateData.lcdbuffer[i][j][0]=' ';
        ThreadStateData.lcdbuffer[i][j][1]='\0';
        ThreadStateData.lcddirty[i][j]=true;
    }
    ThreadStateData.replayString[0] = ' ';
    ThreadStateData.replayString[1] = '\0';
    ThreadStateData.State=Menu;
    for (i=0;i<LCDMAXSTATEBUF;i++) LastState[i]=Title; LastStateP=0;
    ThreadStateData.barx=1, ThreadStateData.bary=1, ThreadStateData.barl=0;
    for (i=0;i<LCDMAXCARDS;i++) ThreadStateData.CardStat[i]=0;
    
    conv = new cCharSetConv(cCharSetConv::SystemCharacterTable() ? cCharSetConv::SystemCharacterTable() : "UTF-8", "ISO-8859-1");
}

cLcd::~cLcd() {
    if (connected) {
        cLcd::Close();
    }
    if (conv) {
        delete(conv);
    }
    if ( femonReceiver ) {
        delete femonReceiver;
        femonReceiver = NULL;
    }
    
}

bool cLcd::Connect( char *host, unsigned int port ) {

    char istring[1024];
    unsigned int i=0;

    if ( ( sock = sock_connect(host,port) ) < 1)
    {
        return connected=false;
    }

    ToggleMode=false;
    sock_send_string(sock, "hello\n");
    usleep(800000); // wait for a connect message  
    sock_recv(sock, istring, 1024);

    if ( strncmp("connect LCDproc",istring,15) != 0 ) {   
        syslog(LOG_INFO, "testing connect LCDproc: %s", istring);
        cLcd::Close();
        return connected=false;
    }

    while ( (strncmp("lcd",istring+i,3) != 0 ) && (i<(strlen(istring)-5)) ) i++;

    if (sscanf( istring+i,"lcd wid %d hgt %d cellwid %d cellhgt %d"
            , &wid, &hgt, &cellwid, &cellhgt) ) connected=true;

    if ((hgt < 4 ) || (wid < 16)) connected = false; // 4 lines are needed, may work with more than 4 though
    if ( (hgt==2) && (wid>31) ) { // 2x32-2x40
        connected = true;
        wid=wid/2;
        LineMode=0;
    }
    else if ( (hgt==2) && (wid>15) && (wid<32) ) { // 2x16-2x31
        connected = true;
        ToggleMode=true;
    }
    if (!connected) {
        cLcd::Close();
        return connected;
    }

    sock_send_string(sock,"screen_add VDR\n"); sock_recv(sock, istring, 1024);
    sock_send_string(sock,"screen_set VDR -heartbeat off\n"); sock_recv(sock, istring, 1024);
    sock_send_string(sock,"widget_add VDR line1 string\n"); sock_recv(sock, istring, 1024);
    sock_send_string(sock,"widget_add VDR line2 string\n"); sock_recv(sock, istring, 1024);
    sock_send_string(sock,"widget_add VDR line3 string\n"); sock_recv(sock, istring, 1024);
    sock_send_string(sock,"widget_add VDR line4 string\n"); sock_recv(sock, istring, 1024);
    sock_send_string(sock,"widget_add VDR prbar hbar\n"); sock_recv(sock, istring, 1024);
    sock_send_string(sock,"widget_set VDR prbar 1 1 0\n"); sock_recv(sock, istring, 1024);

    char cmd[160]; // for the SpAn-Bars
    for(i = 0; i < wid; i++) {
        snprintf(cmd, 160, "widget_add VDR SaBar%d vbar\n", i+1);
        sock_send_string(sock,cmd); sock_recv(sock, istring, 1024);
        snprintf(cmd, 160, "widget_set VDR SaBar%d %d %d 15\n",i+1,i+1,hgt);
        sock_send_string(sock,cmd); sock_recv(sock, istring, 1024);
    }

    sock_send_string(sock,"widget_add VDR SaScroller scroller\n"); sock_recv(sock, istring, 1024);
    SAScroller = true;

    for (i=0; i<LcdMaxKeys;i++) {
        sprintf(istring,"client_add_key %c\n",LcdUsedKeys[i]);
        sock_send_string(sock,istring);
        sock_recv(sock, istring, 1024);
    }

    cLcd::Start(); SetThreadState(Title); // Start thread
    return connected;
}

void cLcd::Close() {
    if ( LcdSetup.isSoundgraphLCD ) {
        char timerText[wid + 1];
        timerText[wid] = '\0';
        cTimer *nextTimer = NULL;
        nextTimer = Timers.GetNextActiveTimer();
        if ( nextTimer ) { // see patches/lcdproc-custom-goodbye-message.diff for details
                           // as the display's backlight always stays on, this way
                           // the next timer is displayed during the time the computer
                           // is turned off (soft-off)
            time_t start = nextTimer->StartTime();
            struct tm times;
            char startBuf[25];
            strftime(startBuf, sizeof(startBuf), "%d.%R", localtime_r(&start, &times));
            snprintf(timerText, wid+1, "%s %s", startBuf, nextTimer->Channel()->ShortName(true) );
            
            cLcd::Write( 1, timerText );
            cLcd::Write( 2, nextTimer->File() );
            
            cSafeFile file = cSafeFile("/etc/LCDd.custom");
            if ( file.Open() ) {
                fprintf( file, "%s\n%s\n", timerText, nextTimer->File() );
                file.Close();
            }
        }
    }
    
    BeginMutualExclusion();
    char istring[1024];
    fprintf(stderr,"vdr-LCDproc: Close Called \n");
    if (connected) {
        char lcdCommand[100];
        closing = true;
        sock_send_string(sock,"output off\n");
        sock_recv(sock, istring, 1024);
        usleep(1000000);
        if ( LcdSetup.isSoundgraphLCD ) {
            ThreadStateData.outputValue = 0;
            ThreadStateData.outputValue |= 1<<IMON_OUTPUT_TIME;
            sprintf(lcdCommand,"output %d\n", ThreadStateData.outputValue);
            usleep(500000);
            sock_send_string(sock,lcdCommand);
            usleep(500000);
        }
        sock_close(sock);
    }
    else {
        fprintf(stderr,"Not Connected !!! \n");
    }
    connected=false; sock=wid=hgt=cellwid=cellhgt=0;
    EndMutualExclusion();
}

const char* cLcd::Convert(const char *s) {
    // do character recoding to ISO-8859-1
    if (!s) return s;
    if (strlen(s)==0) return s;
    const char *s_converted = conv->Convert(s);
    if (s_converted == s) {
        esyslog("lcdproc-plugin: conversion from %s to ISO-8859-1 failed.", cCharSetConv::SystemCharacterTable());
        esyslog("lcdproc-plugin: '%s'",s);
    }
    return s_converted;
}


void cLcd::Info() { // just for testing ...
    if (connected)
        printf("sock %d, wid %d, hgt %d, cellwid %d, cellhgt %d\n",sock, wid, hgt, cellwid, cellhgt);
    else
        printf("not connected\n");
}

void cLcd::SetTitle( const char *string ) {
    if (!connected)
        return;

    const char* c_string = Convert(string);
    
    unsigned int i;
    char title[wid+1];
    // we need to know the length of the translated string "Schedule"
    const char *trstring=trVDR("Schedule - %s");
    int trstringlength=strlen(trstring)-3;

    if (strncmp(trstring,string,trstringlength)==0) {
        // it is the title of the "Schedule" menu, so we replace "Schedule" with ">"
        title[0]='>'; snprintf( title+1, wid,"%s",c_string+trstringlength);
    }
    else if (strlen(string) > (wid-1)) {
        snprintf( title, wid+1,"%s",c_string);
    }
    else {
        memset(title,' ',wid/2+1);
        snprintf(title + ( (wid-strlen(c_string))/2 ),wid-( (wid-strlen(c_string))/2 ),"%s",c_string); // center title
        for (i=0;i<strlen(title);i++) title[i]=toupper(title[i]); // toupper
    }
    cLcd::SetLine(LCDMENU,0,title);
}

void cLcd::SetMain( unsigned int n, const char *string, bool isConverted ) {
    if (!connected)
        return;

    const char* c_string;

    if (isConverted)
        c_string = string;
    else
        c_string = Convert(string);

    char line2[wid+1];
    char line3[wid+1];

    if (c_string != NULL) {
        cLcd::Copy(ThreadStateData.lcdfullbuffer[n],c_string,LCDMAXFULLSTRING-3);
        int i = strlen(ThreadStateData.lcdfullbuffer[n]);
        ThreadStateData.lcdfullbuffer[n][i++]=' ';
        ThreadStateData.lcdfullbuffer[n][i++]='*';
        ThreadStateData.lcdfullbuffer[n][i++]=' ';
        ThreadStateData.lcdfullbuffer[n][i]='\0';
        ThreadStateData.newscroll=true;
        cLcd::Copy(StringBuffer,string,2*wid);
        cLcd::Split(StringBuffer,line2,line3);
        if ( hgt == 2 && n == LCDMISC ) {
            cLcd::SetBuffer(n,line2,line3,NULL,NULL);
        }
        else {
            cLcd::SetBuffer(n,NULL,line2,line3,NULL);
        }
    }
    else {
        //cLcd::SetBuffer(n,NULL," \0"," \0",NULL);
        ThreadStateData.lcdfullbuffer[n][0]='\0';
    }
    
    
    // to store the item that was selected as last for being able to recognize
    // which media-type is going to be played back.
    // i.e.: When selecting a .ogg from mp3-plugin, get the extension and determine
    // which icon to turn on on the Soundgraph-LCD.
    // This is not the best way, but I don't know any better.
    // When a .m3u has been selected we are pretty f***** off ;-)
    if ( !LcdSetup.isSoundgraphLCD )
        return;
    
    if ( n == 0 ) {
        char tmp[4];
        cLcd::Copy( tmp, string + strlen(string) - 3, 3 );

        if  ( strncmp( string," @current", 8 ) && strncmp( tmp,"...", 3 ) ) {
//            cLcd::Copy( ThreadStateData.replayString, string + strlen(string) - 4, 4 ); // just the extension (i.e. ".mpg" or "mpeg")
            cLcd::Copy( ThreadStateData.replayString, string, strlen(string) );
        }
//        syslog(LOG_INFO, "%s\n", ThreadStateData.lcdfullbuffer[LCDREPLAY]);
    }
    
    // ugly hack to get the default-title the playing-plugin sends with the replay-data
    // before it sends the real title - we need the keyword DVD,MP3,...!
    if ( ( replayDvbApi && !strncmp(ThreadStateData.lcdfullbuffer[LCDREPLAY],"DVD", 3 ) ) ||
                ( replayDvbApi && ( ThreadStateData.outputValue & (1 << IMON_OUTPUT_DISK_IN) ) != 0 ) ) // DVD is already playing
    {
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_TOP_ROW); // unset entire row
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BL_ICONS); // unset entire row
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BM_ICONS); // unset entire row
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BR_ICONS); // unset entire row
        ThreadStateData.outputValue |= 4 << IMON_OUTPUT_TOP_ROW; // DVD
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_DISK_IN;
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_DISC;
//        syslog(LOG_INFO, "output DVD - 1");
    }
    if ( replayDvbApi && !strncmp(ThreadStateData.lcdfullbuffer[LCDREPLAY],"MP3", 3 ) ) {
//        syslog(LOG_INFO, "2%s", ThreadStateData.replayString);
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_TOP_ROW); // unset entire row
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BL_ICONS); // unset entire row
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BM_ICONS); // unset entire row
        ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BR_ICONS); // unset entire row
//        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
//        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_BR_ICONS; // MP3
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_DISC;
//        syslog(LOG_INFO, "output MP3 - 1");
        
        cLcd::determineReplayTypeIcons( ThreadStateData.replayString );
    }
}
/**
 * Known problems: When starting a MPlayer replay during playing a VDR recording,
 *                 the replay is not identified - seems to be a problem of the
 *                 standard vdr-lcdproc anyway. Additionally the MPlayer replay
 *                 may have the the wrong aspect ratio at such cases. Thus you
 *                 really do not want to do such a thing ;-).
 *                 On return from a MPlayer replay the audio-tracks seem not to
 *                 be reset correctly, thus the speaker-icons remain the way they
 *                 are (the same can be observed with the femon plugin). 
 */
void cLcd::determineReplayTypeIcons( const char* string )
{
    if ( strlen(string) < 5 ) // avoid invalid strings or strings definetively
        return;               // without any file extension
    
    char extension[5];
    int tmpOutputValue1 = ThreadStateData.outputValue;
    
    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_TOP_ROW); // unset entire row
    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BL_ICONS); // unset entire row
    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BM_ICONS); // unset entire row
    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BR_ICONS); // unset entire row
    
    int tmpOutputValue2 = ThreadStateData.outputValue;
    
    if ( string[ strlen(string) - 1 ] == ' ' && string[strlen(string) - 2 ] == '*' &&
            string[strlen(string) - 3 ] == ' ' )
        cLcd::Copy( extension, string + strlen(string) - 7, 4 ); // just the extension (i.e. ".mpg" or "mpeg")
    else
        cLcd::Copy( extension, string + strlen(string) - 4, 4 ); // just the extension (i.e. ".mpg" or "mpeg")
    
//    syslog(LOG_INFO, "extension = '%s'", string);
//    syslog(LOG_INFO, "extension = '%s'", extension);
    
    if ( strncmp( extension, ".mp3", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_BR_ICONS; // MP3
    }
    else if ( strncmp( extension, ".ogg", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_BR_ICONS; // OGG
    }
    else if ( strncmp( extension, ".wma", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
        ThreadStateData.outputValue |= 3 << IMON_OUTPUT_BR_ICONS; // WMA
    }
    else if ( strncmp( extension, ".wav", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
        ThreadStateData.outputValue |= 3 << IMON_OUTPUT_BR_ICONS; // WAV
    }
    else if ( strncmp( extension, ".mp2", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_BM_ICONS; // MPG
    }
    else if ( strncmp( extension, ".ac3", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_BM_ICONS; // AC3
    }
    else if ( strncmp( extension, ".dts", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
        ThreadStateData.outputValue |= 3 << IMON_OUTPUT_BM_ICONS; // DTS
    }
    else if ( strncmp( extension, ".mpg", 4 ) == 0 ||
            strncmp( extension, "mpeg", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_BL_ICONS; // MPG
    }
    else if ( strncmp( extension, "divx", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_BL_ICONS; // DIVX
    }
    else if ( strncmp( extension, "xvid", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
        ThreadStateData.outputValue |= 3 << IMON_OUTPUT_BL_ICONS; // XVID
    }
    else if ( strncmp( extension, ".wmv", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
        ThreadStateData.outputValue |= 4 << IMON_OUTPUT_BL_ICONS; // WMV
    }
    else if ( strncmp( extension, ".avi", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
    }
    else if ( strncmp( extension, ".mkv", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
    }
    else if ( strncmp( extension, ".264", 4 ) == 0 )
    {
        ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
    }
    // TODO - did I forget anything? Please let me know ;-)
    
    if ( ThreadStateData.outputValue == tmpOutputValue2 )
        ThreadStateData.outputValue = tmpOutputValue1;
}

void cLcd::ClearSaBars() {
    unsigned int i;
    char cmd[80];
    BeginMutualExclusion();
    for(i = 0; i < wid; i++)
    {
        sprintf(cmd, "widget_set VDR SaBar%d %d %d %d\n",i+1,i+1,hgt,0);
        sock_send_string(sock, cmd);
        if ( SAScroller )
        {
            sock_send_string(sock,"widget_del VDR SaScroller\n");
            SAScroller = false;
        }
    }
    EndMutualExclusion();
}

void cLcd::SetHelp( unsigned int n, const char *Red, const char *Green, const char *Yellow, const char *Blue ) {
    if (!connected)
        return;

    char help[2*wid], red[wid+1], green[wid+1], yellow[wid+1], blue[wid+1];
    unsigned int allchars=0, i,j , empty=0, spacewid=1;
    char *longest, *longest1, *longest2;
    
    const char* c_Red = Convert(Red);
    if ( c_Red==NULL || c_Red[0]=='\0' ) {
        empty++; red[0]=' '; red[1]='\0';
    }
    else {
        j=i=0; while ( (i<wid) && (c_Red[i] != '\0') ) {
            if (c_Red[i] !=' ') {
                red[j]=c_Red[i]; j++;
            }
            i++;
        }
        red[j]='\0';
        allchars+=strlen(red);
    }
    
    const char* c_Green  = Convert(Green);
    if ( c_Green==NULL || c_Green[0]=='\0' ) {
        empty++; green[0]=' '; green[1]='\0';
    }
    else {
        j=i=0; while ( (i<wid) && (c_Green[i] != '\0') ) {
            if (c_Green[i] !=' ') {
                green[j]=c_Green[i]; j++;
            }
            i++;
        }
        green[j]='\0';
        allchars+=strlen(green);
    }
    
    const char* c_Yellow = Convert(Yellow);
    if ( c_Yellow==NULL || c_Yellow[0]=='\0' ) {
        empty++; yellow[0]=' '; yellow[1]='\0';
    }
    else {
        j=i=0; while ( (i<wid) && (c_Yellow[i] != '\0') ) {
            if (c_Yellow[i] !=' ') {
                yellow[j]=Yellow[i]; j++;
            }
            i++;
        }
        yellow[j]='\0';
        allchars+=strlen(yellow);
    }
    
    const char* c_Blue   = Convert(Blue);
    if ( c_Blue==NULL || c_Blue[0]=='\0' ) {
        empty++; blue[0]=' '; blue[1]='\0';
    }
    else {
        j=i=0; while ( (i<wid) && (c_Blue[i] != '\0') ) {
            if (c_Blue[i] !=' ') {
                blue[j]=c_Blue[i]; j++;
            }
            i++;
        }
        blue[j]='\0';
        allchars+=strlen(blue);
    }
    while (allchars > (wid-empty-3)) {
        longest1=( strlen(red)      > strlen(green)    ) ? red     : green;
        longest2=( strlen(yellow)   > strlen(blue)     ) ? yellow  : blue;
        longest =( strlen(longest1) > strlen(longest2) ) ? longest1: longest2;
        longest[strlen(longest)-1]='\0'; allchars--;
    }

    if (empty>0) {
        spacewid=(wid-allchars-3)/empty;
        if (spacewid<1) spacewid=1;
        char spacer[spacewid+1]; memset(spacer,'-',spacewid); spacer[spacewid]='\0';
        if ( Red==NULL    || Red[0]=='\0' )    strncpy(red,spacer,spacewid+1);
        if ( Green==NULL  || Green[0]=='\0' )  strncpy(green,spacer,spacewid+1);
        if ( Yellow==NULL || Yellow[0]=='\0' ) strncpy(yellow,spacer,spacewid+1);
        if ( Blue==NULL   || Blue[0]=='\0' )   strncpy(blue,spacer,spacewid+1);
    }

#define MANYBLANKS "%s                                         "

    snprintf(help,wid,MANYBLANKS,red);
    snprintf(help+strlen(red)+1,wid,MANYBLANKS,green);
    snprintf(help+wid-strlen(blue)-strlen(yellow)-1,wid,MANYBLANKS,yellow);
    snprintf(help+wid-strlen(blue),strlen(blue)+1,"%s",blue);
    for (i=0;i<strlen(help);i++) help[i]=toupper(help[i]);
    cLcd::SetLine(n,3,help);
}

void cLcd::SetStatus( const char *string ) {
    if (!connected)
        return;

    const char* c_string = Convert(string);
    
    char statstring[2*wid+1];

    if (c_string == NULL) {
        cLcd::SetMain(LCDMENU,StringBuffer, true);
    }
    else {
        cLcd::Copy(statstring,c_string,2*wid);
        cLcd::SetMain(LCDMENU,statstring);
    }
}

void cLcd::SetWarning( const char *string ) {
    if (!connected)
        return;
    
    const char* c_string = Convert(string);

    char statstring[2*wid+1];

    if (c_string != NULL) {
        cLcd::Copy(statstring,c_string,2*wid);
        cLcd::Clear(LCDMISC);
        cLcd::SetMain(LCDMISC,statstring,true);
        cLcd::SetThreadState(Misc);
    }
}

void cLcd::ShowVolume( unsigned int vol, bool absolute ) {
    if (!connected)
        return;
    
    BeginMutualExclusion();
    if (absolute)
        ThreadStateData.volume=vol;
    else
        ThreadStateData.volume+=vol;
    ThreadStateData.muted=(ThreadStateData.volume==0);
    
//    ThreadStateData.volume=primaryDvbApi->CurrentVolume();
//    ThreadStateData.muted=primaryDvbApi->IsMute();
    ThreadStateData.showvolume=true;
    EndMutualExclusion();
    if (ThreadStateData.muted) {
        if ( hgt == 2 ) {
            cLcd::SetLine(Vol,0,Convert(tr("Mute")));
            cLcd::SetLine(Vol,1," ");
        }
        else {
            cLcd::SetLine(Vol,0," ");
            cLcd::SetLine(Vol,1,Convert(tr("Mute")));
            cLcd::SetLine(Vol,2," ");
            cLcd::SetLine(Vol,3," ");
        }
    }
    else {
        if (hgt==2) {
            if (ToggleMode) {
                cLcd::SetLine(Vol,0,"|---|---|---|---|---|---|---|---|---|---");
                cLcd::SetLine(Vol,1," ");
            }
            else {
                cLcd::SetLine(Vol,0,"|---|---|---|---|---|---|---|---|---|---");
                cLcd::SetLine(Vol,3," ");
                cLcd::SetLine(Vol,1,"|---|---|---|---|---|---|---|---|---|---");
                cLcd::SetLine(Vol,2," ");
            }
        }
        else {
            cLcd::SetLine(Vol,0,Convert(tr("Volume ")));
            cLcd::SetLine(Vol,3,"|---|---|---|---|---|---|---|---|---|---");
            cLcd::SetLine(Vol,1,"|---|---|---|---|---|---|---|---|---|---");
            cLcd::SetLine(Vol,2," ");
        }
    }
}

void cLcd::SetProgress( const char *begin, const char *end, int percent ) {
    if (!connected)
        return;

    char workstring[256];

    if (percent>100) percent=100;
    else if (percent<0) percent=0;
    if (begin==NULL) {
        BeginMutualExclusion();
            ThreadStateData.barx=1; ThreadStateData.bary=1; ThreadStateData.barl=0;
        EndMutualExclusion();
    }
    else {
        time_t t = time(NULL);
        if (t != LastProgress) { // output only once a second
            int beginw=strlen(begin);
            int endw=strlen(end);
            sprintf(workstring,"%s", begin);
            memset(workstring+beginw,' ',wid-beginw-endw);
            sprintf(workstring+wid-endw,"%s", end);
            cLcd::SetLine(LCDREPLAY,(ToggleMode)?0:3,workstring);
            BeginMutualExclusion();
            ThreadStateData.barx=beginw+1+((hgt==2 && !ToggleMode)?wid:0);
            ThreadStateData.bary=((hgt==2)?1:4);
            ThreadStateData.barl=(percent*cellwid*(wid-beginw-endw))/100;
            if ( LcdSetup.isSoundgraphLCD )
                ThreadStateData.progressBarsValue = cLcd::progressBar( ThreadStateData.progressBarsValue, percent/(double)100 );
            EndMutualExclusion();
            LastProgress = t;
        }
    }
}

void cLcd::SetLine( unsigned int n, unsigned int l, const char *string ) {
    if (!connected)
        return;
    
    BeginMutualExclusion();
        if (string != NULL) strncpy(ThreadStateData.lcdbuffer[n][l],string,wid+1);
        ThreadStateData.lcddirty[n][l]=true;
    EndMutualExclusion();
}

void cLcd::SetLineC( unsigned int n, unsigned int l, const char *string ) {
    if (!connected)
        return;
    
    const char* c_string = Convert(string);

    BeginMutualExclusion();
        if (c_string != NULL) cLcd::Copy(ThreadStateData.lcdbuffer[n][l],c_string,wid);
        ThreadStateData.lcddirty[n][l]=true;
    EndMutualExclusion();
}

void cLcd::SetBuffer( unsigned int n, const char *l1,const char *l2,const char *l3,const char *l4 ) {
    if (!connected)
        return;

    BeginMutualExclusion();
        if (l1 != NULL) strncpy(ThreadStateData.lcdbuffer[n][0],l1,wid+1); ThreadStateData.lcddirty[n][0]=true;
        if (l2 != NULL) strncpy(ThreadStateData.lcdbuffer[n][1],l2,wid+1); ThreadStateData.lcddirty[n][1]=true;
        if (l3 != NULL) strncpy(ThreadStateData.lcdbuffer[n][2],l3,wid+1); ThreadStateData.lcddirty[n][2]=true;
        if (l4 != NULL) strncpy(ThreadStateData.lcdbuffer[n][3],l4,wid+1); ThreadStateData.lcddirty[n][3]=true;
    EndMutualExclusion();
}

void cLcd::SetRunning( bool nownext, const char *string1, const char *string2, const char *string3 ) {
    if (!connected)
        return;

    char line[1024];
    char line1[1024];

    static char now1[LCDMAXWID+1];
    static char now2[LCDMAXWID+1];

    char *c_string1 = string1 ? strdup(Convert(string1)) : NULL;
    char *c_string2 = string2 ? strdup(Convert(string2)) : NULL;
    char *c_string3 = string3 ? strdup(Convert(string3)) : NULL;
    
    snprintf(line,1024,"%s %s%s%s",
        (string1==NULL || string1[0]=='\0')?" ":c_string1,
        (string2==NULL || string2[0]=='\0')?" ":c_string2,
        (string3==NULL || string3[0]=='\0')?" ":"|",
        (string3==NULL || string3[0]=='\0')?" ":c_string3);
    cLcd::Copy(line1,line,2*wid);
    
    free(c_string1);
    free(c_string2);
    free(c_string3);

    
    if (nownext) {
        BeginMutualExclusion();
            cLcd::Split(line1,ThreadStateData.lcdbuffer[LCDMISC][2],ThreadStateData.lcdbuffer[LCDMISC][3]);
            ThreadStateData.lcddirty[LCDMISC][2]=true; ThreadStateData.lcddirty[LCDMISC][3]=true;
        EndMutualExclusion();
        cLcd::SetBuffer(LCDMISC,now1,now2,NULL,NULL);
    }
    else {
        BeginMutualExclusion();
            cLcd::Split(line1,ThreadStateData.lcdbuffer[LCDTITLE][2],ThreadStateData.lcdbuffer[LCDTITLE][3]);
            ThreadStateData.lcddirty[LCDTITLE][2]=true; ThreadStateData.lcddirty[LCDTITLE][3]=true;
            cLcd::Copy(ThreadStateData.lcdfullbuffer[LCDTITLE],line,LCDMAXFULLSTRING-3);
            int i = strlen(ThreadStateData.lcdfullbuffer[LCDTITLE]);
            ThreadStateData.lcdfullbuffer[LCDTITLE][i++]=' ';
            ThreadStateData.lcdfullbuffer[LCDTITLE][i++]='*';
            ThreadStateData.lcdfullbuffer[LCDTITLE][i++]=' ';
            ThreadStateData.lcdfullbuffer[LCDTITLE][i]='\0';
            sprintf(now1,"%s",ThreadStateData.lcdbuffer[LCDTITLE][2]);
            sprintf(now2,"%s",ThreadStateData.lcdbuffer[LCDTITLE][3]);
        EndMutualExclusion();
    }
}

void cLcd::SummaryInit( const char *string ) {
    if (SummaryText)
        free(SummaryText);
    SummaryText  = strdup(Convert(string));
    SummaryTextL = strlen(SummaryText);
    SummaryCurrent=0;
}

void cLcd::SummaryDown() {
    if (!connected)
        return;

    SummaryCurrent=( (SummaryCurrent+hgt*wid) < SummaryTextL )?SummaryCurrent+hgt*wid:SummaryCurrent;
}

void cLcd::SummaryUp() {
    if (!connected) return;

    SummaryCurrent=( SummaryCurrent > hgt*wid )?SummaryCurrent-hgt*wid:0;
}

void cLcd::SummaryDisplay() {
    if ( (!connected) || (SummaryTextL < 1) )
        return;

    char workstring[256];
    unsigned int i, offset=0;

    for (i=0;i<hgt;i++) {
        if (SummaryCurrent+offset < SummaryTextL) {
            strncpy(workstring,SummaryText+SummaryCurrent+offset,wid+1);
            workstring[wid]='\0';
            cLcd::SetLine(LCDMISC,i,workstring);
            offset+=wid;
        }
        else cLcd::SetLine(LCDMISC,i," ");
    }
}

void cLcd::SetCardStat( unsigned int card, unsigned int stat ) {
    if (!connected)
        return;
    BeginMutualExclusion();
        ThreadStateData.CardStat[card]=stat;
    EndMutualExclusion();
}

void cLcd::Clear( unsigned int n ) {
    if (!connected)
        return;

    cLcd::SetBuffer(n," "," "," "," ");
}

void cLcd::SetThreadState( ThreadStates newstate ) {
    if (!connected)
        return;

    BeginMutualExclusion();
        if (ThreadStateData.State != newstate)
        {
            cLcd::LastState[(++LastStateP)%LCDMAXSTATEBUF]=newstate;
            ThreadStateData.State=newstate;
            ThreadStateData.barx=1, ThreadStateData.bary=1, ThreadStateData.barl=0;
    
            unsigned int i;
            char cmd[80];
            for(i = 0; i < wid; i++)
            {
                sprintf(cmd, "widget_set VDR SaBar%d %d %d %d\n",i+1,i+1,hgt,0);
                sock_send_string(sock, cmd);
            }
        }
    EndMutualExclusion();
}

void cLcd::PopThreadState() {
    BeginMutualExclusion();
        LastStateP=(LastStateP+LCDMAXSTATEBUF-1)%LCDMAXSTATEBUF;
        ThreadStateData.State=cLcd::LastState[LastStateP];
        ThreadStateData.barx=1, ThreadStateData.bary=1, ThreadStateData.barl=0;
    EndMutualExclusion();
}

void cLcd::SetReplayDevice( cControl *DvbApi ) {
    replayDvbApi=DvbApi;
}

void cLcd::SetPrimaryDevice( cDevice *DvbApi ) {
    primaryDvbApi=DvbApi;
}

// private:

// everything between BeginMutualExclusion(); and EndMutualExclusion();
// should have exclusive access to ThreadStateData

void cLcd::BeginMutualExclusion() {
    CriticalArea.Lock();
}

void cLcd::EndMutualExclusion() {
    cLcd::CriticalArea.Unlock();
}

void cLcd::Copy( char *to, const char *from, unsigned int max ) { // eliminates tabs, double blanks ...

    unsigned int i=0 , out=0;

    if (from != NULL) {
        while ((out < max) && (from[i] != '\0') )         {
            to[out]=(isspace(from[i]))?' ':from[i];
            if ( (out>0) && (to[out-1]==' ') && ispunct(to[out]) ) {
                to[out-1]=to[out]; to[out]=' ';
            }
            if (!(( (out==0) && (to[out]==' ') ) ||
                    ( (out>0) && (to[out-1]==' ') && (to[out]==' ') ) ||
                    ( (out>0) && ispunct(to[out-1]) && (to[out]==' ') )
            )) out++;
            i++;
        }
        if (out==0) to[out++]=' ';
        to[out]='\0';
    }
    else {
        to[0]=' '; to[1]='\0';
    }
}

void cLcd::Split( char *string, char *string1, char *string2 ) {

    unsigned int i,j,k,ofs;

    if ( hgt>2 && ( strlen(string) < 2*wid) && isdigit(string[0]) && isdigit(string[1]) // beautification ...
            && string[2]==':' && isdigit(string[3]) && isdigit(string[4]) ) {
        char *tmpptr=strpbrk(string,"|");
        if ( ( tmpptr != NULL ) && (ofs=tmpptr-string)<wid && wid-(ofs+1)<2*wid-(j=strlen(string)) ) {
            ofs=wid-(ofs+1);
            string[j+ofs]='\0';
            for (i=j+ofs-1; i>=wid-ofs;i--) string[i]=string[i-ofs];
            for (i=wid-ofs-1;i<wid;i++) string[i]=' ';
        }
    }

    if ( hgt>2 && ( strlen(string) < 2*wid) && ( strlen(string) >= wid ) ) { // beautification ...

        if (isalpha(string[wid-1]) && isalpha(string[wid]) ) {
            j=strlen(string);
            for (i=wid-1; (i>0) && (string[i]!=' ') && (string[i]!='|') && (string[i]!=',') && (string[i]!=')')
            && (string[i]!='-') && (string[i]!='.') && (string[i]!=':'); i-- )
            {}

            if ( ( (2*wid-j) >= (ofs=wid-(i+1)) ) && ofs+j <= 2*wid ) {
                string[j+ofs]='\0';
                for (k=j+ofs-1;k>i+ofs; k-- ) string[k]=string[k-ofs];
                for (k=0;k<ofs;k++) string[i+k+1]=' ';
            }

        }

        if ( (j=strlen(string)) < 2*wid && isdigit(string[0]) && isdigit(string[1]) &&
                string[2]==':' && isdigit(string[3]) && isdigit(string[4]) ) {
            ofs=2*wid-strlen(string); ofs=(ofs>6)?6:ofs;
            if (string[wid]=='|') string[wid]=' ';
            if (ofs==6 && string[wid]==' ') ofs--;
            string[j+ofs]='\0';
            for (i=j+ofs-1; i>=wid+ofs;i--) string[i]=string[i-ofs];
            for (i=wid;i<wid+ofs;i++) string[i]=' ';
        }
    }

    strncpy(string1,string,wid+1);
    if ( strlen(string) >wid ) {
        strncpy(string2,string+wid,wid+1);
    }
    else {
        string2[0]=' '; string2[1]='\0';
    }
}

void cLcd::Write( int line, const char *string ) { // used for any text output to LCDd

    char workstring[256];
    unsigned int i,out;

    if (ToggleMode && (line > 2))
        return;

    if (ToggleMode || hgt>2 ) {
        sprintf(workstring,"widget_set VDR line%d 1 %d \"",line,line);
    }
    else if (LineMode==0) {
        sprintf(workstring,"widget_set VDR line%d %d %d \"",line,(line==2||line==4)?wid+1:1,(line<3)?1:2 );
    }
    else {
        sprintf(workstring,"widget_set VDR line%d %d %d \"",line,(line==3||line==4)?wid+1:1,(line==1||line==4)?1:2);
    }
    out=strlen(workstring);
    for (i=0;(i<strlen(string)) && (i<wid);i++)
        workstring[out++] = LcdTransTbl[LcdSetup.Charmap][ (unsigned char) string[i] ]; // char mapping see lcdtranstbl.h
    workstring[out++] = '"'; workstring[out++] = '\n'; workstring[out] = '\0';
    BeginMutualExclusion();
        sock_send_string(sock,workstring);
    EndMutualExclusion();
}

void cLcd::GetTimeDateStat( char *string, unsigned int OutStateData[] ) {
    time_t t;
    struct tm *now;
    unsigned int i;
    const char States[] = ".-*>"; // '.' ... not present, '-' ... present, '*' ... recording, '>' ... replaying
    bool ShowStates=false;
    unsigned int offset;

    offset=(wid>36)?20:0;
    for (i=0; i<LCDMAXCARDS && (!ShowStates); i++) ShowStates = ShowStates || (OutStateData[i]>1);

    t = time(NULL);
    now = localtime(&t);

    if ( offset || !( ShowStates && ((t%LcdSetup.FullCycle) >= LcdSetup.TimeCycle) )) {
        if (wid > 19)
            snprintf(string,wid+1,"<%s %02d.%02d %02d:%02d:%02d>",
                    Convert(*WeekDayName(now->tm_wday)), now->tm_mday, now->tm_mon+1, now->tm_hour, now->tm_min,now->tm_sec);
        else
            snprintf(string,wid+1,"<%02d.%02d %02d:%02d:%02d>",
                    now->tm_mday, now->tm_mon+1, now->tm_hour, now->tm_min,now->tm_sec);
    }

    if ( offset || ( ShowStates && ((t%LcdSetup.FullCycle) >= LcdSetup.TimeCycle) )) {
        if (LcdSetup.RecordingStatus == 0) {
            for (i=0; i<LCDMAXCARDS; i++) {
                snprintf(string+offset,5," %d:%c", i,States[ OutStateData[i] ] );
                offset+=4;
            }
        }
        else {
            snprintf(string,wid+1,"<%s %02d:%02d:%02d>", Convert(tr("RECORDING")), now->tm_hour, now->tm_min,now->tm_sec);
        }
    }
}

inline int cLcd::volumeBar( int oldValue, double newValue ) {
    oldValue &= ~0x00FC0000;
    oldValue |= ((int)( 32 * newValue ) << 18) & 0x00FC0000;
    oldValue |= 1<<IMON_OUTPUT_PBARS;
    return oldValue;
}

inline int cLcd::progressBar( int oldValue, double newValue ) {
    oldValue &= ~0x00000FC0;
    oldValue |= ((int)( 32 * newValue ) << 6) & 0x00000FC0;
    oldValue |= 1<<IMON_OUTPUT_PBARS;
    return oldValue;
}

inline bool cLcd::isIMonLCD(void) {
    if ( !connected )
        return false;
    
    char istring[1024];

    sock_send_string(sock,"info\n");
    usleep(500000); // wait for the answer  
    sock_recv(sock, istring, 1024);
    
    if ( strstr( istring, "SoundGraph iMON OEM (and others) LCD driver" ) )
        return true;
    
    return false;
}

void cLcd::audioTrackChanged() {
    if (femonReceiver) {
        delete femonReceiver;
        femonReceiver = NULL;
    }
    attachFemonReceiver = true;
}

#define WakeUpCycle 125000 // us 
void cLcd::Action(void) { // LCD output thread
    unsigned int i,j, barx=1, bary=1, barl=0, ScrollState=0, ScrollLine=1, lasttitlelen=0;
    int Current=0, Total=1, scrollpos=0, scrollcnt=0, scrollwaitcnt=10, lastAltShift=0, lastBackLight,lastPrio, lastPrioN, keycnt=0;
    struct timeval now, voltime;
    char workstring[1024], lastkeypressed='\0';
    cLcd::ThreadStates PrevState=Menu;
    struct cLcd::StateData OutStateData;
    bool Lcddirty[LCDMAXSTATES][4];
    bool LcdShiftkeyPressed=false;
    int progressBarsValueOld = 0;
    int outputValueOld = 0;
    attachFemonReceiver = true;    
    gettimeofday(&lastIcons,NULL);
    
    if (primaryDvbApi) for (int k=0; k<primaryDvbApi->NumDevices(); k++) SetCardStat(k,1);
    
    if ( cLcd::isIMonLCD() )
        LcdSetup.isSoundgraphLCD = 1;
    else
        LcdSetup.isSoundgraphLCD = 0;
    
    char priostring[35];
    // RT
    static int rtcycle;
    // LCR
    static int lcrCycle;

    // backlight init 
    if ((lastBackLight=LcdSetup.BackLight)) {
        sock_send_string(sock,"backlight on\n");
    }
    else {
        sock_send_string(sock,"backlight off\n");
    }

    // prio init
    if ( (lastPrio=LcdSetup.SetPrio) ) {
        if (LcdSetup.SetPrio == 1)
            sprintf(priostring,"screen_set VDR -priority %d\n", LcdSetup.ClientPrioN );
        else
            sprintf(priostring,"screen_set VDR -priority %d\n", LcdSetup.ClientPrioH );
        sock_send_string(sock,priostring);
    }
    lastPrioN=LcdSetup.ClientPrioN;

    syslog(LOG_INFO, "LCD output thread started (pid=%d), display size: %dx%d", getpid(),hgt,wid);
    if ( hgt == 4 ) {
        cLcd::Write(1," Welcome  to  V D R\0");
        cLcd::Write(2,"--------------------\0");
        cLcd::Write(3,"Video Disk Recorder\0");
        cLcd::Write(4,"Version: "VDRVERSION"\0");
    }
    else if ( hgt == 2 ) {
        cLcd::Write(1,"Welcome to VDR\0");
        cLcd::Write(2,"Version: "VDRVERSION"\0");
    }

    // Output init
    if (LcdSetup.OutputNumber > 0) {
        char lcdCommand[100];

//        for (int o=0; o < LcdSetup.OutputNumber; o++)
//        {
//            sprintf(lcdCommand,"output %d\n",1 << o);
//            sock_send_string(sock,lcdCommand);
//            usleep(150000);
//        }
//        for (int o= LcdSetup.OutputNumber-1; o >= 0; o--)
//        {
//            sprintf(lcdCommand,"output %d\n",1 << o);
//            sock_send_string(sock,lcdCommand);
//            usleep(150000);
//        }
        
        sprintf(lcdCommand,"output on\n");
        sock_send_string(sock,lcdCommand);
        usleep(500000);
        sprintf(lcdCommand,"output off\n");
        sock_send_string(sock,lcdCommand);
    }

    sleep(3); bool volume=false; OutStateData.showvolume=false; ThreadStateData.showvolume=false;
    
    if ( primaryDvbApi ) {
        BeginMutualExclusion();
            ThreadStateData.volume=primaryDvbApi->CurrentVolume();
            ThreadStateData.muted=primaryDvbApi->IsMute();
            if ( LcdSetup.isSoundgraphLCD ) {
                ThreadStateData.progressBarsValue = cLcd::volumeBar( ThreadStateData.progressBarsValue,
                            ThreadStateData.volume / (double)255 );
                sprintf( workstring,"output %d\n", ThreadStateData.progressBarsValue );
                sock_send_string( sock, workstring );
            }
        EndMutualExclusion();
        usleep(WakeUpCycle);
    }
    
    //TODO: Check the following line
    if (primaryDvbApi) for (int k=0; k<primaryDvbApi->NumDevices() ; k++) SetCardStat(k,1);
    
    voltime.tv_sec=0;
    for (i=0;i<LCDMAXSTATES;i++) for (j=0;j<4;j++) Lcddirty[i][j]=true;
    //  time_t nextLcdUpdate = (time(NULL)/60)*60+60;

    time_t nextLcdUpdate = 0; // trigger first update immediately

//    gettimeofday(&dimm_time,NULL);
//    dimmed = false;
    while ( Running() && !closing && connected ) { // main loop, wakes up every WakeUpCycle, any output to LCDd is done here  
        gettimeofday(&now,NULL);

        //  epg update
        if (channelSwitched) {
            channelSwitched = false;
            nextLcdUpdate = 0; //trigger next epg update
            attachFemonReceiver = true;
//            backlightOn( &dimm_time );
        }


        if ( time(NULL) > nextLcdUpdate && !replayDvbApi ) {
//            cChannel *channel = Channels.GetByNumber(primaryDvbApi->CurrentChannel());
            cChannel *channel = Channels.GetByNumber(cDevice::CurrentChannel());
            const cEvent *Present = NULL;
            cSchedulesLock schedulesLock;
            const cSchedules *Schedules = cSchedules::Schedules(schedulesLock);
            if ( Schedules && channel) {
                const cSchedule *Schedule = Schedules->GetSchedule(channel->GetChannelID());
                if (Schedule) {
                    const char *PresentTitle, *PresentSubtitle;
                    PresentTitle = NULL; PresentSubtitle = NULL;
                    if ((Present = Schedule->GetPresentEvent()) != NULL) {
                        if ( LcdSetup.isSoundgraphLCD ) {
                            BeginMutualExclusion();
                            // funny side effect: When the program overruns, the bar
                            // extends from right to left - nice, since it was unintended
                            // but pretty useful though ;-)
                            ThreadStateData.progressBarsValue = cLcd::progressBar( ThreadStateData.progressBarsValue,
                                    ( time(NULL) - Present->StartTime() ) / (double) Present->Duration() );
                            EndMutualExclusion();
    
                            // for progress bar:
                            nextLcdUpdate=(time(NULL))+10;
                        }
                        else {
                            nextLcdUpdate=Present->StartTime()+Present->Duration();
                        }
                        
                        PresentTitle = Present->Title();
                        PresentSubtitle = Present->ShortText();
                        if ( (!isempty(PresentTitle)) && (!isempty(PresentSubtitle)) )
                            SetRunning(false,Present->GetTimeString(),PresentTitle,PresentSubtitle);
                        else if (!isempty(PresentTitle)) SetRunning(false,Present->GetTimeString(),PresentTitle);
                    }
                    else
                        SetRunning(false,tr("No EPG info available."), NULL);
                    if ((Present = Schedule->GetFollowingEvent()) != NULL)
                        nextLcdUpdate=(Present->StartTime()<nextLcdUpdate)?Present->StartTime():nextLcdUpdate;
                        rtcycle = 10; // RT
                        lcrCycle = 10; // LCR
                }
            }
            if ( nextLcdUpdate <= time(NULL) )
                nextLcdUpdate=(time(NULL)/60)*60+60;
        }


        // get&display Radiotext
        if (++rtcycle > 10) { // every 10 times
            cPlugin *p;
            p = cPluginManager::CallFirstService("RadioTextService-v1.0", NULL);
            if (p) {
                RadioTextService_v1_0 rtext;
                if (cPluginManager::CallFirstService("RadioTextService-v1.0", &rtext)) {
                    if (rtext.rds_info == 2 && strstr(rtext.rds_title, "---") == NULL) {
                        char timestr[20];
                        sprintf(timestr, "%02d:%02d", rtext.title_start->tm_hour, rtext.title_start->tm_min);
                        SetRunning(false, timestr, rtext.rds_title, rtext.rds_artist);
                    }
                    else if (rtext.rds_info > 0) {
                        SetRunning(false, NULL, rtext.rds_text);
                    }
                }
            }
            rtcycle = 0;
            //printf("lcdproc - get Radiotext ...\n");
        }

        // get&display LcrData
        if (lcrCycle++ == 10) { // every 10 times
            lcrCycle = 0;
            cPlugin *p;
            p = cPluginManager::CallFirstService("LcrService-v1.0", NULL);
            if (p) {
                LcrService_v1_0 lcrData;
                if (cPluginManager::CallFirstService("LcrService-v1.0", &lcrData)) {
                    if ( strstr( lcrData.destination, "---" ) == NULL ) {
                        SetRunning(false, (const char *)lcrData.destination, (const char *)lcrData.price, (const char *)lcrData.pulse);
                        nextLcdUpdate = 0; //trigger next epg update
                    }
                }
            }
        }


        // replaying

        //    if ( (now.tv_usec < WakeUpCycle) && (replayDvbApi) ) {
        //    if ( (now.tv_usec < WakeUpCycle/8) && (replayDvbApi) ) {
        if ( (replayDvbApi) ) {
//            backlightOn( &dimm_time );
            if ( ThreadStateData.State == SA ) {
                BeginMutualExclusion();
                    ThreadStateData.State = OutStateData.State = PrevState = Replay;
                EndMutualExclusion();
            }
            
            char tempbuffer[16];
            if ( replayDvbApi ) {
                replayDvbApi->GetIndex(Current, Total, false);
                Total=(Total==0)?1:Total;
                sprintf(tempbuffer,IndexToHMSF(Total));
                SetProgress(IndexToHMSF(Current),tempbuffer, (100 * Current) / Total);
            }

            if ( LcdSetup.enableSpectrumAnalyzer && ThreadStateData.State != LCDMENU
                    && !ThreadStateData.showvolume && cPluginManager::CallFirstService(SPAN_GET_BAR_HEIGHTS_ID, NULL)){
                
                Span_GetBarHeights_v1_0 GetBarHeights;
                char cmd[160];
                unsigned int *barHeights = new unsigned int[wid];
                unsigned int *barHeightsLeftChannel = new unsigned int[wid];
                unsigned int *barHeightsRightChannel = new unsigned int[wid];
                unsigned int volumeLeftChannel;
                unsigned int volumeRightChannel;
                unsigned int volumeBothChannels;
                const char *saClient = "lcdproc";
                unsigned int *barPeaksBothChannels = new unsigned int[wid];
                unsigned int *barPeaksLeftChannel = new unsigned int[wid];
                unsigned int *barPeaksRightChannel = new unsigned int[wid];

                GetBarHeights.bands = wid;
                GetBarHeights.barHeights = barHeights;
                GetBarHeights.barHeightsLeftChannel = barHeightsLeftChannel;
                GetBarHeights.barHeightsRightChannel = barHeightsRightChannel;
                GetBarHeights.volumeLeftChannel = &volumeLeftChannel;
                GetBarHeights.volumeRightChannel = &volumeRightChannel;
                GetBarHeights.volumeBothChannels = &volumeBothChannels;
                GetBarHeights.name = saClient;
                GetBarHeights.falloff = 12;
                GetBarHeights.barPeaksBothChannels = barPeaksBothChannels;
                GetBarHeights.barPeaksLeftChannel = barPeaksLeftChannel;
                GetBarHeights.barPeaksRightChannel = barPeaksRightChannel;

                if ( cPluginManager::CallFirstService(SPAN_GET_BAR_HEIGHTS_ID, &GetBarHeights )) {                    
                    BeginMutualExclusion();
                        OutStateData.State = ThreadStateData.State = SA;
                        for(i = 0; i < wid; i++) {
                            sprintf(cmd, "widget_set VDR SaBar%d %d %d %d\n",i+1,i+1,hgt,((cellhgt*(hgt/2)/*1 or 2 lines high*/)*barHeights[i])/100);
                            sock_send_string(sock, cmd);
                        }
                    EndMutualExclusion();
                }
                delete [] barHeights;
                delete [] barHeightsLeftChannel;
                delete [] barHeightsRightChannel;
                delete [] barPeaksBothChannels;
                delete [] barPeaksLeftChannel;
                delete [] barPeaksRightChannel;
            }

            if ( OutStateData.State != SA && SAScroller == true ) {
                sock_send_string(sock,"widget_del VDR SaScroller\n");
                SAScroller = false;
            }
        }

        // copy 

        BeginMutualExclusion(); // all data needed for output are copied here  
            memcpy(&OutStateData,&ThreadStateData, sizeof (cLcd::StateData));
            ThreadStateData.showvolume=false;
            if (ThreadStateData.newscroll)
            {   scrollpos=0; scrollwaitcnt=LcdSetup.Scrollwait; ThreadStateData.newscroll=false;}
            for (i=0;i<LCDMAXSTATES;i++) for (j=0;j<4;j++)
            {
                ThreadStateData.lcddirty[i][j]=false;
                Lcddirty[i][j]= Lcddirty[i][j] || OutStateData.lcddirty[i][j];
            }
        EndMutualExclusion();

        // scroller

        if ( (OutStateData.State==PrevState) && ( OutStateData.State == Replay || OutStateData.State == Menu || OutStateData.State == Title ) ) {
            switch (OutStateData.State) {
            case Replay:
                ScrollState=LCDREPLAY; ScrollLine=1;
                break;
            case Menu:
                ScrollState=LCDMENU; ScrollLine=1;
                break;
            case Title:
                if (!ToggleMode) {
                    ScrollState=LCDTITLE; ScrollLine=2;
                }
                else {
                    ScrollState=LCDTITLE; ScrollLine=1;
                    char tmpbuffer[1024];
                    strcpy(tmpbuffer,OutStateData.lcdbuffer[LCDTITLE][1]);
                    strcat(tmpbuffer," * ");
                    strcat(tmpbuffer, OutStateData.lcdfullbuffer[ScrollState]);
                    strcpy(OutStateData.lcdfullbuffer[ScrollState],tmpbuffer);
                    strncpy(OutStateData.lcdbuffer[LCDTITLE][1],OutStateData.lcdfullbuffer[ScrollState],wid);
                }
                if ( strlen(OutStateData.lcdfullbuffer[ScrollState]) != lasttitlelen ) {
                    lasttitlelen=strlen(OutStateData.lcdfullbuffer[ScrollState]);
                    scrollpos=0; scrollwaitcnt=LcdSetup.Scrollwait; ThreadStateData.newscroll=false;
                }
                break;
            default:
                break;
            }

            if ( ( strlen(OutStateData.lcdfullbuffer[ScrollState]) > (((ToggleMode)?1:2)*wid+3) )
                    && ( (scrollpos) || !(scrollwaitcnt=(scrollwaitcnt+1)%LcdSetup.Scrollwait) ) ) {
                if ( !(scrollcnt=(scrollcnt+1)%LcdSetup.Scrollspeed) ) {
                    scrollpos=(scrollpos+1)%strlen(OutStateData.lcdfullbuffer[ScrollState]);
                    if ( scrollpos==1 ) scrollwaitcnt=0;
                    for (i=0; i<wid; i++) {
                        OutStateData.lcdbuffer[ScrollState][ScrollLine][i]=
                            OutStateData.lcdfullbuffer[ScrollState][(scrollpos+i)%strlen(OutStateData.lcdfullbuffer[ScrollState])];
                    }
                    OutStateData.lcdbuffer[ScrollState][ScrollLine][wid]='\0';
                    for (i=0; i<wid; i++) {
                        OutStateData.lcdbuffer[ScrollState][ScrollLine+1][i]=
                            OutStateData.lcdfullbuffer[ScrollState][(scrollpos+wid+i)%strlen(OutStateData.lcdfullbuffer[ScrollState])];
                    }
                    OutStateData.lcdbuffer[ScrollState][ScrollLine+1][wid]='\0';
                    Lcddirty[ScrollState][ScrollLine]=Lcddirty[ScrollState][ScrollLine+1]=true;
                }
            }
        }

        // volume  

        if (OutStateData.showvolume)
            gettimeofday(&voltime,NULL);
        if ( voltime.tv_sec != 0) { // volume
            if ( ((now.tv_sec - voltime.tv_sec)*1000000+now.tv_usec-voltime.tv_usec ) > (100000*LcdSetup.VolumeKeep) ) {
                voltime.tv_sec=0;
                OutStateData.barx=1; OutStateData.bary=1; OutStateData.barl=0;
                volume=false;
                
                if ( LcdSetup.isSoundgraphLCD ) {
                    BeginMutualExclusion();
                        ThreadStateData.outputValue &= ~( 1 << IMON_OUTPUT_VOL );
                    EndMutualExclusion();
                }
            }
            else {
                volume=true;
                OutStateData.barx=1; OutStateData.bary=((hgt==2)?2:3);
                OutStateData.barl=(cellwid*wid*OutStateData.volume)/255;
//                backlightOn( &dimm_time );
                if ( LcdSetup.isSoundgraphLCD ) {
                    BeginMutualExclusion();
                        ThreadStateData.progressBarsValue = cLcd::volumeBar( ThreadStateData.progressBarsValue,
                                OutStateData.volume / (double)255 );
                        ThreadStateData.outputValue |= 1<<IMON_OUTPUT_VOL;
                    EndMutualExclusion();
                }
            }
        }
        if (volume) {
            cLcd::ClearSaBars(); // get rid off the bars when volume shall be shown
            OutStateData.State = Vol;
        }

        // prio


        // modes

        switch (OutStateData.State) {

        case Menu: // display menu = 0
//            backlightOn( &dimm_time );
            LineMode=1; if (!volume)
            {   OutStateData.barx=1; OutStateData.bary=1; OutStateData.barl=0;}
            if (PrevState != Menu) for (i=0;i<4;i++) Lcddirty[LCDMENU][i]=true;
            for (i=0;i<4;i++) if (Lcddirty[LCDMENU][i]) {
                cLcd::Write(i+1,OutStateData.lcdbuffer[LCDMENU][i]);
                Lcddirty[LCDMENU][i]=false;
            }
            PrevState=Menu;
            break;

        case Title: // Display 'titlescsreen' = 1
            LineMode=0;
            if ( (now.tv_usec < WakeUpCycle) || (PrevState != Title) ) {
                cLcd::GetTimeDateStat(workstring,OutStateData.CardStat);
                cLcd::Write(1,workstring);

            }
            if (PrevState != Title) for (i=1;i<4;i++) Lcddirty[LCDTITLE][i]=true;
            for (i=1;i<4;i++) if (Lcddirty[LCDTITLE][i]) {
                cLcd::Write(i+1,OutStateData.lcdbuffer[LCDTITLE][i]);
                Lcddirty[LCDTITLE][i]=false;
            }

            PrevState = Title;
            break;

        case Replay: // Display date/time during replaying = 2
            LineMode=1;
            if ( hgt == 2 )
                OutStateData.bary = 1;
            else
                OutStateData.bary = hgt;
            if ( !ToggleMode && ((now.tv_usec < WakeUpCycle) || (PrevState != Replay)) ) {
                cLcd::GetTimeDateStat(workstring,OutStateData.CardStat);
                cLcd::Write(1,workstring);
            }
            if (PrevState != Replay)
            {   scrollpos=0; for (i=1;i<4;i++) Lcddirty[LCDREPLAY][i]=true;}
            for (i=(ToggleMode)?0:1;i<4;i++) if (Lcddirty[LCDREPLAY][i]) {
                cLcd::Write(i+1,OutStateData.lcdbuffer[LCDREPLAY][i]);
                Lcddirty[LCDREPLAY][i]=false;
            }

            PrevState = Replay;
            break;

        case SA: // display the SA, move the progressbar to top and show the title on just one line (scrolling)
            LineMode=0;

            if ( PrevState != SA ) {
                sprintf(replayTextTmp, " "); // triggers e.g. a menu-display right before
            }
            sprintf(workstring,"%s",OutStateData.lcdfullbuffer[LCDREPLAY]);

            if ( hgt > 1 && strcasecmp(workstring,replayTextTmp) != 0 ) {
                if ( SAScroller ) {
                    sock_send_string(sock,"widget_del VDR SaScroller\n");
                    sock_send_string(sock,"widget_add VDR SaScroller scroller\n");
                }
                else {
                    sock_send_string(sock,"widget_add VDR SaScroller scroller\n");
                }
                SAScroller = true;
                sprintf(replayTextTmp,"%s",OutStateData.lcdfullbuffer[LCDREPLAY]);

                if ( hgt == 2 )
                    sprintf(workstring,"widget_set VDR SaScroller %d %d %d %d m 1 \"%s\"\n",1,1,wid,1,replayTextTmp);
                else if ( hgt == 4 )
                    sprintf(workstring,"widget_set VDR SaScroller %d %d %d %d m 3 \"%s\"\n",1,2,wid,2,replayTextTmp);
                sock_send_string(sock,workstring);
            }
            BeginMutualExclusion();
                Lcddirty[LCDREPLAY][0]=false;
                if ( hgt == 4 ) {
                    cLcd::Write(1,OutStateData.lcdbuffer[LCDREPLAY][3]);
                    cLcd::Write(3," ");
                    cLcd::Write(4," ");
                    Lcddirty[LCDREPLAY][1]=false;
                    Lcddirty[LCDREPLAY][2]=false;
                    Lcddirty[LCDREPLAY][3]=false;
                }
                else if ( hgt == 2 ) {
//                    cLcd::Write(1," ");
                    cLcd::Write(2," ");
                    Lcddirty[LCDREPLAY][1]=false;
                    Lcddirty[LCDREPLAY][2]=false;
                }
                OutStateData.bary = 1;
            EndMutualExclusion();

            PrevState = SA;
            break;

        case Misc: // Display messages  = 3
//            backlightOn( &dimm_time );
            LineMode=0;
            if (PrevState != Misc) for (i=0;i<4;i++) Lcddirty[LCDMISC][i]=true;
            for (i=0;i<4;i++) if (Lcddirty[LCDMISC][i]) {
                cLcd::Write(i+1,OutStateData.lcdbuffer[LCDMISC][i]);
                Lcddirty[LCDMISC][i]=false;
            }
            PrevState = Misc;
            break;

        case Vol: // Display Volume = 4
//            backlightOn( &dimm_time );
            LineMode=0;
            if (PrevState != Vol) for (i=0;i<4;i++) Lcddirty[LCDVOL][i]=true;
            for (i=0;i<4;i++) if (Lcddirty[LCDVOL][i]) {
                cLcd::Write(i+1,OutStateData.lcdbuffer[LCDVOL][i]);
                Lcddirty[LCDVOL][i]=false;
            }
          
            PrevState = Vol;
            break;

        default: // quite impossible :) 
            break;
        }

        // bargraph
        if ( !(OutStateData.State == SA && hgt == 2 && OutStateData.barl > 0)
                && !(OutStateData.State == Replay && hgt == 2 && OutStateData.barl > 0)
                && !((OutStateData.State == Menu || OutStateData.State == Title) && LcdSetup.isSoundgraphLCD && OutStateData.barl > 0)
                && ( (OutStateData.barx != barx) || (OutStateData.bary != bary) || (OutStateData.barl != barl) ) ) {
            sprintf(workstring,"widget_set VDR prbar %d %d %d\n",OutStateData.barx,OutStateData.bary,OutStateData.barl);
            sock_send_string(sock,workstring);
            barx=OutStateData.barx; bary=OutStateData.bary; barl=OutStateData.barl;
        }

        // backlight
        BeginMutualExclusion();
            if ( lastBackLight != LcdSetup.BackLight ) {
                lastBackLight=LcdSetup.BackLight;
                if (lastBackLight) {
                    sock_send_string(sock,"backlight on\n");
                }
                else {
                    sock_send_string(sock,"backlight off\n");
                }
            }
        EndMutualExclusion();
        // keys

        if ( LcdMaxKeys && (lastAltShift != LcdSetup.AltShift) ) {
            lastAltShift=LcdSetup.AltShift;
            if (lastAltShift)
                sock_send_string(sock,"screen_set VDR -heartbeat slash\n");
            else
                sock_send_string(sock,"screen_set VDR -heartbeat heart\n");
        }

        if ( !(keycnt=(keycnt+1)%4) ) lastkeypressed='\0';

        workstring[0]='\0'; sock_recv(sock, workstring, 1023); workstring[1024]='\0';
        if ( LcdMaxKeys && ( strlen(workstring) > 4 ) ) {
            for (i=0; i < (strlen(workstring)-4); i++ ) {
                if (workstring[i]=='k' && workstring[i+1]=='e' && workstring[i+2]=='y'
                    && workstring[i+3]==' ' && workstring[i+4]!=lastkeypressed ) {
                    lastkeypressed=workstring[i+4];
                    for (j=0; j<LcdMaxKeys && workstring[i+4]!=LcdUsedKeys[j]; j++ )
                    {}
                    if (workstring[i+4] == LcdShiftKey) {
                        LcdShiftkeyPressed = ! LcdShiftkeyPressed;
                        if (LcdShiftkeyPressed)
                            sock_send_string(sock,"screen_set VDR -heartbeat on\n");
                        else
                            sock_send_string(sock,"screen_set VDR -heartbeat off\n");
                    }
                    if ( (workstring[i+4] != LcdShiftKey) ) {
                        if (LcdShiftkeyPressed) {
                            //syslog(LOG_INFO, "shiftkey  pressed: %c %d",  workstring[i+4],j);
                            cRemote::Put(LcdShiftMap[j]);
                            LcdShiftkeyPressed=false;
                            sock_send_string(sock,"screen_set VDR -heartbeat off\n");
                        }
                        else {
                            //syslog(LOG_INFO, "normalkey pressed: %c %d",  workstring[i+4],j);
                            cRemote::Put(LcdNormalMap[j]);
                        }
                    }
                }
            }
        }

        // Output

        int OutValue = 0;
        char lcdCommand[100];
        if ( !closing && !LcdSetup.isSoundgraphLCD ) {
            if (LcdSetup.OutputNumber > 0){
                for (int o=0; o <  LcdSetup.OutputNumber; o++) {
                    switch(LcdSetup.OutputFunction[o]) {           
                    case 0: // Off                   
                        break;
                    case 1: // On
                        OutValue += 1 << o;                                                                               
                        break;               
                    case 2: // Recording DVB 1
                        if (OutStateData.CardStat[0] == 2)
                            OutValue += 1 << o;             
                        break;
                    case 3: // Recording DVB 2
                        if (OutStateData.CardStat[1] == 2)
                            OutValue += 1 << o;
                        break;                                                                                                
                    case 4: // Recording DVB 3
                        if (OutStateData.CardStat[2] == 2)
                            OutValue += 1 << o;             
                        break;
                    case 5: // Recording DVB 4
                        if (OutStateData.CardStat[3] == 2)
                            OutValue += 1 << o;             
                        break;
                    case 6: // Replay
                        if (OutStateData.State == Replay)
                            OutValue += 1 << o;            
                        break;
                    case 7: // DVD
                        if ( (OutStateData.State == Replay) && (!strncmp(OutStateData.lcdfullbuffer[LCDREPLAY],"DVD", 3)))
                            OutValue += 1 << o;                                                                             
                        break;               
                    case 8: // Mplayer
                        break;
                    case 9: // MP3
                        if ( (OutStateData.State == Replay) && OutStateData.lcdfullbuffer[LCDREPLAY][0]=='[' &&
                                OutStateData.lcdfullbuffer[LCDREPLAY][3]==']' && 
                                (OutStateData.lcdfullbuffer[LCDREPLAY][1]=='.' || OutStateData.lcdfullbuffer[LCDREPLAY][1]=='L' ) &&
                                (OutStateData.lcdfullbuffer[LCDREPLAY][2]=='.' || OutStateData.lcdfullbuffer[LCDREPLAY][2]=='S' )   
                        )
                            OutValue += 1 << o;
                        break;
                    case 10: // Mplayer + MP3
                        // Until I find a better solution any replay that is not a DVD is flagged as Mplayer-Mp3
                        if ( (OutStateData.State == Replay) && (strncmp(OutStateData.lcdfullbuffer[LCDREPLAY],"DVD",3))!= 0)
                            OutValue += 1 << o;
                        break;
                    case 11: // User1
                        break;
                    case 12: // User2
                        break;
                    case 13: // User3
                        break;
                    default:
                        ;
                    }  
                }    
                sprintf(lcdCommand,"output %d\n",OutValue);
                BeginMutualExclusion();
                    sock_send_string(sock,lcdCommand);
                EndMutualExclusion();
            }
        }
        /**
         * Sets the "output state" for the device. We use this to control the icons around the outside the
         * display. The bits in \c state correspond to the icons as follows:
         *
         * bit 0       : disc icon (0=off, 1='spin') , if Toprow==4, use CD-animation, else use "HDD-recording-animation"
         * bit 1,2,3   : top row (0=none, 1=music, 2=movie, 3=photo, 4=CD/DVD, 5=TV, 6=Web, 7=News/Weather)
         * bit 4,5     : 'speaker' icons (0=off, 1=L+R, 2=5.1ch, 3=7.1ch)
         * bit 6       : S/PDIF icon
         * bit 7       : 'SRC'
         * bit 8       : 'FIT'
         * bit 9       : 'TV'
         * bit 10      : 'HDTV'
         * bit 11      : 'SRC1'
         * bit 12      : 'SRC2'
         * bit 13,14,15: bottom-right icons (0=off, 1=MP3, 2=OGG, 3=WMA, 4=WAV)
         * bit 16,17,18: bottom-middle icons (0=off, 1=MPG, 2=AC3, 3=DTS, 4=WMA)
         * bit 19,20,21: bottom-left icons (0=off, 1=MPG, 2=DIVX, 3=XVID, 4=WMV)
         * bit 22      : 'VOL' (volume)
         * bit 23      : 'TIME'
         * bit 24      : 'ALARM'
         * bit 25      : 'REC' (recording)
         * bit 26      : 'REP' (repeat)
         * bit 27      : 'SFL' (shuffle)
         * bit 28      : Abuse this for progress bars (if set to 1), lower bits represent
         *               the length (6 bits each: P|6xTP|6xTL|6xBL|6xBP with P = bit 28,
         *               TP=Top Progress, TL = Top Line, BL = Bottom Line, BP = Bottom Progress).
         *               If bit 28 is set to 1, lower bits are interpreted as
         *               lengths; otherwise setting the symbols as usual.
         *               0 <= length <= 32, bars extend from left to right.
         *               length > 32, bars extend from right to left, length is counted
         *               from 32 up (i.e. 35 means a length of 3).
         *                       
         *     Remember: There are two kinds of calls!
         *               With bit 28 set to 1: Set all bars (leaving the symbols as is),
         *               with bit 28 set to 0: Set the symbols (leaving the bars as is).
         *     Beware:   TODO: May become a race condition, if both calls are executed
         *                     before the display gets updated. Keep this in mind in your
         *                     client-code.
         * bit 29      : 'disc-in icon' - half ellipsoid under the disc symbols (0=off, 1=on)
         * TODO: Entire structure may change on future releases of the iMon-driver, so
         *       check it from time to time.
         */
        else if ( !closing ) {
            BeginMutualExclusion();

                ThreadStateData.outputValue &= ~(1 << IMON_OUTPUT_SFL);
                ThreadStateData.outputValue &= ~(1 << IMON_OUTPUT_REP);
                ThreadStateData.outputValue &= ~(1 << IMON_OUTPUT_ALARM);
    
                if ( PrevState == Misc )
                    ThreadStateData.outputValue |= 1 << IMON_OUTPUT_ALARM;
    
                if ( replayDvbApi ) {
                    ThreadStateData.outputValue &= ~( 3<<IMON_OUTPUT_SPKR );
                    ThreadStateData.outputValue &= ~( 1<<IMON_OUTPUT_SPDIF );
                }
    
    
                if ( ( replayDvbApi && !strncmp(OutStateData.lcdfullbuffer[LCDREPLAY],"DVD", 3 ) ) ||
                        ( replayDvbApi && ( ThreadStateData.outputValue & (1 << IMON_OUTPUT_DISK_IN) ) != 0 ) ) { // DVD is already playing
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_TOP_ROW); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BL_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BM_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BR_ICONS); // unset entire row
                    ThreadStateData.outputValue |= 4 << IMON_OUTPUT_TOP_ROW; // DVD
                    ThreadStateData.outputValue |= 1 << IMON_OUTPUT_DISK_IN;
                    ThreadStateData.outputValue |= 1 << IMON_OUTPUT_DISC;
                    //                syslog(LOG_INFO, "output DVD2");
                }
                else if ( ( replayDvbApi && OutStateData.lcdfullbuffer[LCDREPLAY][0]=='[' &&
                        OutStateData.lcdfullbuffer[LCDREPLAY][3]==']' && 
                        (OutStateData.lcdfullbuffer[LCDREPLAY][1]=='.' || OutStateData.lcdfullbuffer[LCDREPLAY][1]=='L' ) &&
                        (OutStateData.lcdfullbuffer[LCDREPLAY][2]=='.' || OutStateData.lcdfullbuffer[LCDREPLAY][2]=='S' ) ) ||
                        ( replayDvbApi && !strncmp(ThreadStateData.lcdfullbuffer[LCDREPLAY],"MP3", 3 ) &&
                                (  replayDvbApi && ( ThreadStateData.outputValue & (1 << IMON_OUTPUT_TOP_ROW) ) != 0 || 
                                        replayDvbApi && ( ThreadStateData.outputValue & (2 << IMON_OUTPUT_TOP_ROW) ) != 0 )
                        )
                ) {
                    // everything else already has been set (hopefully)
                    if ( OutStateData.lcdfullbuffer[LCDREPLAY][1]=='L' )
                        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_REP;
                    if ( OutStateData.lcdfullbuffer[LCDREPLAY][2]=='S' )
                        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_SFL;
                    //                syslog(LOG_INFO, "output MP3 - 3");
                }
                else if ( ( replayDvbApi && OutStateData.lcdfullbuffer[LCDREPLAY][0]=='[' &&
                        OutStateData.lcdfullbuffer[LCDREPLAY][3]==']' && 
                        (OutStateData.lcdfullbuffer[LCDREPLAY][1]=='.' || OutStateData.lcdfullbuffer[LCDREPLAY][1]=='L' ) &&
                        (OutStateData.lcdfullbuffer[LCDREPLAY][2]=='.' || OutStateData.lcdfullbuffer[LCDREPLAY][2]=='S' ) ) ||
                        replayDvbApi && !strncmp(ThreadStateData.lcdfullbuffer[LCDREPLAY],"MP3", 3 )
    
                ) {
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_TOP_ROW); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BL_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BM_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BR_ICONS); // unset entire row
                    ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // Music
                    ThreadStateData.outputValue |= 1 << IMON_OUTPUT_BR_ICONS; // MP3
                    ThreadStateData.outputValue |= 1 << IMON_OUTPUT_DISC;
                    //                syslog(LOG_INFO, "output MP3 - 2");
                }
                else if ( replayDvbApi ) {
                    // standard VDR recording
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_TOP_ROW); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BL_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BM_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BR_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(1 << IMON_OUTPUT_DISK_IN);
                    ThreadStateData.outputValue |= 2 << IMON_OUTPUT_TOP_ROW; // Movie
                    ThreadStateData.outputValue |= 1 << IMON_OUTPUT_DISC;
    
                    // let's see, if there is a MPlayer-replay
                    determineReplayTypeIcons( ThreadStateData.lcdfullbuffer[LCDREPLAY] );
    
                    // do a rather rudimental audio detection, compared to the one
                    // we can use during live-tv
                    eTrackType track = cDevice::ActualDevice()->GetCurrentAudioTrack();
                    if ( IS_AUDIO_TRACK(track) && !IS_DOLBY_TRACK(track) ) {
                        if ( cDevice::ActualDevice()->GetAudioChannel() == 0 ) // stereo
                            ThreadStateData.outputValue |= 1<<IMON_OUTPUT_SPKR;
                        else if ( cDevice::ActualDevice()->GetAudioChannel() == 1 || // mono left
                                cDevice::ActualDevice()->GetAudioChannel() == 2 ) // mono right
                            ThreadStateData.outputValue &= ~( 1<<IMON_OUTPUT_SPKR );
    
                        ThreadStateData.outputValue &= ~( 1<<IMON_OUTPUT_SPDIF );
                    }
                    else if ( IS_DOLBY_TRACK(track) ) {
                        ThreadStateData.outputValue |= 1<<IMON_OUTPUT_SPDIF;
                        ThreadStateData.outputValue |= 1<<IMON_OUTPUT_SPKR; // stereo
                    }
    
                    //                syslog(LOG_INFO, "output VDR-Recording or mplayer-replay");
                }
                else { // no replay, this should be TV or radio
                       // see cLcd::determineReplayTypeIcons() for known problems, if you're searching them here
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_TOP_ROW); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BL_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BM_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(7 << IMON_OUTPUT_BR_ICONS); // unset entire row
                    ThreadStateData.outputValue &= ~(1 << IMON_OUTPUT_DISK_IN);
                    ThreadStateData.outputValue &= ~(1 << IMON_OUTPUT_DISC);
                    ThreadStateData.outputValue &= ~(3<<IMON_OUTPUT_SPKR);
                    ThreadStateData.outputValue &= ~(1<<IMON_OUTPUT_SPDIF);
    
                    cChannel *channel = Channels.GetByNumber(cDevice::CurrentChannel());
                    if ( channel != NULL && channel->Vpid() == 0 && channel->Apid(0) > 0 ) {
                        ThreadStateData.outputValue |= 1 << IMON_OUTPUT_TOP_ROW; // radio (=music)
                        //                    syslog(LOG_INFO, "radio VDR-output");
                    }
                    else {
                        ThreadStateData.outputValue |= 5 << IMON_OUTPUT_TOP_ROW; // TV is standard
                        //                    syslog(LOG_INFO, "tv VDR-output");
                    }
    
                    // TODO: check the following in productive operation, might lead to a VDSB
                    // when returning from a replay and a timer got startet in the meantime
                    // which revoke the channel that had been tuned to prior the start of the replay.
                    // at least on my old machine - my new computer is not affected!
                    eTrackType track;
                    if ( cDevice::ActualDevice() != NULL ) {
                        track = cDevice::ActualDevice()->GetCurrentAudioTrack();
                        
                        // needed for budget-only systems,
                        // as the ActualDevice() then seems to get "lost" immediately
                        // after the switch to a channel
                        if ( !track )
                            track = cDevice::PrimaryDevice()->GetCurrentAudioTrack();
                    }
                    
                    if ( attachFemonReceiver && cDevice::ActualDevice() != NULL &&
                            channel != NULL && primaryDvbApi->IsTunedToTransponder(channel) ) {
                        int apid[2] = {0, 0};
                        int dpid[2] = {0, 0};

                        if (femonReceiver) {
                            delete femonReceiver;
                            femonReceiver = NULL;
                        }
                        IS_AUDIO_TRACK(track) ? apid[0] = channel->Apid(int(track - ttAudioFirst)) : apid[0] = channel->Apid(0);
                        IS_DOLBY_TRACK(track) ? dpid[0] = channel->Dpid(int(track - ttDolbyFirst)) : dpid[0] = channel->Dpid(0);
    
                        if ( IS_DOLBY_TRACK(track) && LcdSetup.detectAC3Mode ) {
                            femonReceiver = new cFemonReceiver( channel->GetChannelID(), channel->Ca(), channel->Vpid(), apid, dpid);
                            cDevice::ActualDevice()->ForceTransferMode();
                            cDevice::ActualDevice()->AttachReceiver( femonReceiver );
//                            syslog(LOG_INFO, "cDevice::ActualDevice()->ForceTransferMode()");
                        }
    
                        attachFemonReceiver = false;
                    }// TODO: End of "to check"
                    
                    if ( IS_DOLBY_TRACK(track) ) {
//                                            syslog(LOG_INFO, "dolby track");
                        ThreadStateData.outputValue |= 1<<IMON_OUTPUT_SPDIF;
                        if ( femonReceiver && femonReceiver->AC3_2_0() )
                            ThreadStateData.outputValue |= 1<<IMON_OUTPUT_SPKR;
                        else if ( femonReceiver && femonReceiver->AC3_5_1() )
                            ThreadStateData.outputValue |= 2<<IMON_OUTPUT_SPKR;
                        else
                            // as default assume AC3-Stereo
                            ThreadStateData.outputValue |= 1<<IMON_OUTPUT_SPKR;
                    }
                    else if ( IS_AUDIO_TRACK(track) ) {
                        if ( cDevice::ActualDevice() != NULL && cDevice::ActualDevice()->GetAudioChannel() == 0 ) // stereo
                        {
                            ThreadStateData.outputValue |= 1<<IMON_OUTPUT_SPKR;
//                                                    syslog(LOG_INFO, "stereo track");
                        }
                        else if ( cDevice::ActualDevice() != NULL && ( cDevice::ActualDevice()->GetAudioChannel() == 1 || // mono left
                                cDevice::ActualDevice()->GetAudioChannel() == 2 ) ) { // mono right
                            ThreadStateData.outputValue &= ~( 1<<IMON_OUTPUT_SPKR );
//                                                    syslog(LOG_INFO, "mono track");
                        }
                        ThreadStateData.outputValue &= ~( 1<<IMON_OUTPUT_SPDIF );
                    }
    
                    //                syslog(LOG_INFO, "output VDR-TV");
                }
    
                if ( OutStateData.CardStat[0] == 2 || OutStateData.CardStat[1] == 2
                        || OutStateData.CardStat[2] == 2 || OutStateData.CardStat[3] == 2 ) {
                    ThreadStateData.outputValue |= 1<<IMON_OUTPUT_REC;
                }
                else {
                    ThreadStateData.outputValue &= ~(1<<IMON_OUTPUT_REC);
                }
    
                // for alternating setting the symbols or progress bars
                if ( ThreadStateData.outputValue != outputValueOld || (now.tv_sec - lastIcons.tv_sec)*1000000 + (now.tv_usec - lastIcons.tv_usec) > 1000000*2/*=2 seconds*/ ) {
                    sprintf(lcdCommand,"output %d\n",ThreadStateData.outputValue);
                    outputValueOld = ThreadStateData.outputValue;
//                    syslog(LOG_INFO, "outputValue");
//                    syslog(LOG_INFO, "outputValue %d", ThreadStateData.outputValue);
                    sock_send_string(sock,lcdCommand);
                    usleep(WakeUpCycle + WakeUpCycle/2); // TODO: /2?
                    gettimeofday(&lastIcons,NULL);
                }
                else if ( ThreadStateData.progressBarsValue != progressBarsValueOld ) {
                    sprintf(lcdCommand,"output %d\n",ThreadStateData.progressBarsValue);
                    progressBarsValueOld = ThreadStateData.progressBarsValue;
    //                syslog(LOG_INFO, "progressBarsValue");
                    sock_send_string(sock,lcdCommand);
                    usleep(WakeUpCycle + WakeUpCycle/2);
                    gettimeofday(&lastIcons,NULL);
                }
            EndMutualExclusion();
        }

        backlightOn( /*&dimm_time*/ );
        
//        f = fopen("/video/commands/tmp/vdr-lcd-backlight-on", "r"); // open file for reading
//        if ( f != NULL ) {
//            fclose(f);
//            f = NULL;
//            if ( !dimmed && LcdSetup.BackLightWait > 0 && now.tv_sec - dimm_time.tv_sec > LcdSetup.BackLightWait ) {
//                BeginMutualExclusion();
////                SystemExec( "/video/commands/lcd_control dimm" );
////                usleep( 50000 );
//                    dimmed = true;
//                    gettimeofday(&dimm_time,NULL);
//                EndMutualExclusion();
//            }
//        }
//        else if ( lastBackLight ) {
//            BeginMutualExclusion();
//                LcdSetup.BackLight = 0;
//    //            SystemExec( "/video/commands/lcd_control off" );
//                sock_send_string(sock,"backlight off\n");
//            EndMutualExclusion();
//        }

        usleep(WakeUpCycle-(now.tv_usec%WakeUpCycle)); // sync to systemtime for nicer time output
    }
}


void cLcd::ChannelSwitched() {
    channelSwitched = true;
}

/*
 * For triggering the backlight state externally (e.g. by a script).
 * */
void cLcd::backlightOn( /*struct timeval *dimm_time*/ ) {
    // just for my sweet Matrix-Orbital display, I implemented a dimming mechanism
    // maybe this issue already has been resolved with lcdproc version 0.5 and higher.
    // That's why we might just return here ...
//        return;

//    if ( LcdSetup.BackLightWait > 0 ) // 0 means, that dimming is disabled
//    if ( LcdSetup.BackLight == 2 ) // 0 = off, 1 = on, 2 = auto
    {
        //        if ( dimmed )
        {
            // the script can create or delete the file that indicates,
            // whether the backlight shall be switched on
            // on vdr-start, the file is created automatically (by my vdr-start-script).
            // That way I can see whether the vdr was restarted (bad weather conditions, ...)
            // immediately. When I'm not in front of the vdr-PC, I turn off the backlight.
            f = fopen("/video/commands/tmp/vdr-lcd-backlight-on", "r"); // open file for reading
            if ( f != NULL ) { // check if file exists
                fclose(f);
                f = NULL;
                BeginMutualExclusion();
//                    SystemExec( "/video/commands/lcd_control on" );
//                    sock_send_string(sock,"backlight on\n");
                    LcdSetup.BackLight = 1;
                    // dimmed = false;
                EndMutualExclusion();
            }
            else {
                BeginMutualExclusion();
//                    SystemExec( "/video/commands/lcd_control off" );
//                    sock_send_string(sock,"backlight off\n");
                    LcdSetup.BackLight = 0;
                EndMutualExclusion();
            }
//            dimmed = false;
        }
//        gettimeofday(dimm_time,NULL);
    }
}

#endif //___LCD_C
